from .rules import RC, RGC, Rule, RuleGroup
from .specific import InRange, InSet, IsOptional, IsRequired, IsType, MaxCount, MinCount
from .validate import validate

__all__ = [
    "RC",
    "RGC",
    "Rule",
    "RuleGroup",
    "validate",
    "IsOptional",
    "IsRequired",
    "InRange",
    "MaxCount",
    "MinCount",
    "IsType",
    "InSet",
]



from typing import Any, Callable

from pydantic import BaseModel, Field, ValidationError, create_model, field_validator
from result import Err

from .rules import Rule, RuleGroup
from .specific import IsType


def create_pydantic_model(model_name: str, v_map: dict[str, Any]) -> type[BaseModel]:
    field_definitions: dict[str, Any] = {}
    field_validators: dict[str, classmethod] = {}

    # Add type checks for the model and index validators
    for k, r_rg in v_map.items():
        rules: Rule | RuleGroup | tuple[Rule] = (
            r_rg if isinstance(r_rg, RuleGroup) else Rule.init_specific(r_rg)
        )
        if isinstance(rules, Rule):
            rules = (rules,)
        typ = next((r._type for r in rules if isinstance(r, IsType)), Any)
        field_definitions[k] = (typ, Field(description=f"dyn generated from found `IsType`"))

        non_type_rules = [rule for rule in rules if not isinstance(rule, IsType)]
        if non_type_rules:
            dyn_field_validator = _as_classmethod(non_type_rules)
            field_validators[f"{k}_validator"] = field_validator(k)(dyn_field_validator)  # type: ignore

    model = create_model(model_name, **field_definitions, __validators__=field_validators)  # type: ignore
    return model


def _as_classmethod(rl: list[Rule | RuleGroup]) -> Callable:
    """
    Takes the given list, and wraps it as a class function that just runs the list

    This needs to exist so we have the `cls` parameter as a member function,
        and to raise a pydantic `ValidationError` to actually work correctly
    """
    return lambda cls, d: _check_err(cls, rl, d) or d  # type: ignore


def _check_err(cls, rl: list[Rule | RuleGroup], rl_input: Any) -> None:
    """
    For the given rules, tries to run them. If an `Err` is returned, raise an error
    """
    for r in rl:
        res = r(rl_input)
        if isinstance(res, Err):
            errors = [
                {
                    "loc": (
                        f" ... for the dynamically-generated {cls}",
                    ),  # Use a tuple for location
                    "msg": f"Error for {res} on dynamic class {cls}",
                    "type": "value_error",
                    "input": rl_input,
                    "ctx": {"error": str(res)},  # Add context with 'error' key
                }
            ]
            raise ValidationError.from_exception_data("Dynamic class validation error", errors)  # type: ignore
    return None



from typing import Any, Callable, Mapping

from result import Err, Ok

from ..dicts import get
from .rules import Rule, RuleGroup


def validate(
    source: dict[str, Any],
    validation_map: Mapping[str, Callable | dict[str, Any] | list[Any]],
) -> Ok[dict | list] | Err[list[tuple]]:
    """
    Performs valiudation on the `source` dict. Enforces corresponding `Rule`s and `RuleGroup`s
      at a given key.

    NOTE: This _does_ mutate the corresponding `validation_map` (specifically adds info to the
      `_iter_over` field of Rule | RuleGroup), so it's _not_ a pure function.
    """
    # Try applying each rule at the given key
    failed_r_rg: list[tuple] = []
    for k, v in validation_map.items():
        # Run rules on source[k], return `Err` if get fails
        curr_source = get(source, k, default=None)
        match v:
            case Rule() | RuleGroup():
                res = v(curr_source)
            case dict() | list():
                if curr_source is None:
                    raise RuntimeError(f"Failed to process key {k} from: {source}")
                # Do recursive call. Expect this to return Ok/Err
                if isinstance(v, list):
                    # TODO: this needs to pass a dict, make sure the call here is correct
                    v_dict = {k: v}
                    res = validate(curr_source, v_dict)  # type: ignore
                else:
                    res = validate(curr_source, v)  # type: ignore
            case _:
                if callable(v):
                    # Wrap in a `Rule` so it returns Ok/Err
                    res = Rule.init_specific(v)(curr_source)
                else:
                    raise TypeError(f"Expected `v` to be Callable or nested dict, got: {type(v)}")

        if isinstance(res, Err):
            err_tup: tuple = res.err_value
            failed_r_rg.append(err_tup)

    if failed_r_rg:
        return Err(failed_r_rg)

    return Ok(source)



import ast
import re
from collections.abc import Iterable
from typing import Any, Callable, Literal

import polars as pl
from result import Err

COMMAS_OUTSIDE_OF_BRACKETS = r",(?![^{}\[\]]*[}\]])"
COLON_OUTSIDE_OF_BRACKETS = r":(?![^{]*})"
PERIOD_UP_TO_NEXT_CLOSE_PARENS = r"\.(.*?\))"


def select(
    source: pl.DataFrame,
    key: str,
    consume: bool = False,
    rename: dict[str, str] | Callable[[str], str] | None = None,
) -> pl.DataFrame | Err:
    """
    Selects a subset of a DataFrame. `key` has some convenience functions

    `key` notes:
    - "*" -- all columns
    - "a, b, c" -- columns a, b, c (in-order)
    - "a, b : c > 3" -- columns a, b where column c > 3
    - "* : c != 3" -- all columns where column c != 3
    - "dict_col -> [a, b, c]" -- "dict_col.a, dict_col.b, dict_col.c"
    - "dict_col +> [a, b, c]" -- "dict_col, dict_col -> [a, b, c]"
    - "dict_col -> {"A": a, "B": b}" --"dict_col.a, dict_col.b" and rename `a -> A, b -> B`
    - "dict_col +> {"A": a, "B": b}" -- "dict_col, dict_col -> {"A": a, "B": b}"

    `consume` attempts to drop columns that matched in the `select` from the source dataframe

    `rename` is the standard Polars API call and is called at the very end
    """
    _check_assumptions(source)

    # Extract query from key (if present)
    key = key.replace(" ", "")
    query: pl.Expr | None = None
    if re.search(COLON_OUTSIDE_OF_BRACKETS, key):
        key, query_str = re.split(COLON_OUTSIDE_OF_BRACKETS, key, maxsplit=1)
        query_str = query_str.strip("[]")
        query = _convert_to_polars_filter(query_str)

    # Extract columns from syntax
    parsed_col_list = re.split(COMMAS_OUTSIDE_OF_BRACKETS, key)
    parsed_nested_col_list = _generate_nesting_list(parsed_col_list)

    # Grab correct subset/slice of the dataframe
    try:
        if isinstance(query, pl.Expr):
            source = source.filter(query)
        res = _apply_nested_col_list(source, parsed_nested_col_list)
        # Post-processing checks
        if res.is_empty():
            raise pl.exceptions.ColumnNotFoundError
        if consume:
            # TODO: handle columns with query syntax -- make sure those aren't skipped
            for col_name in parsed_col_list:
                if col_name in source.columns:
                    source.drop_in_place(col_name)
    except pl.exceptions.ColumnNotFoundError:
        return Err("Default Err: `select` key didn't match")

    # TODO: Consider supporting regex search and pattern replacements (e.g. prefix_* -> new_prefix_*)
    if rename and isinstance(res, pl.DataFrame):
        res = res.rename(rename)

    return res


def join(
    source: pl.DataFrame,
    second: pl.DataFrame,
    how: Literal["inner", "left", "cross", "anti", "semi"],
    on: str | list[str],
) -> pl.DataFrame | Err:
    try:
        _pre_merge_checks(source, second, on)
    except KeyError as e:
        return Err(f"Failed pre-merge checks for {how} join: {str(e)}")

    res = source.join(second, how=how, on=on, join_nulls=False, coalesce=True)

    if how == "left":
        # If there were no matches, then return `Err`
        #  Check for non-null cols after the left-join
        matched = True
        for col_name in second.columns:
            matched = matched and res.filter(pl.col(col_name).is_not_null()).height > 0
        if not matched:
            return Err("No matching columns on left join")

    return res if not res.is_empty() else Err("Empty dataframe after left join")


def union(
    source: pl.DataFrame,
    rows=pl.DataFrame | list[dict[str, Any]],
    na_default: Any = None,
) -> pl.DataFrame | Err:
    """
    Inserts rows into the end of the DataFrame

    For a row, if a value is not specified it will be filled with the specified default

    If the union operation cannot be done (e.g. incompatible columns), returns `Err`
    """
    if isinstance(rows, list):
        rows = pl.DataFrame(rows)

    # Ensure all columns in `into` are present in `rows`
    for col in source.columns:
        if col not in rows.columns:
            rows = rows.with_columns(pl.lit(na_default).alias(col))

    # Ensure all columns in `rows` are present in `into`
    for col in rows.columns:
        if col not in source.columns:
            source = source.with_columns(pl.lit(na_default).alias(col))

    try:
        res = pl.concat([source, rows])
    except Exception as e:
        return Err(f"Error when unioning: {str(e)}")

    return res


def group_by(source: pl.DataFrame, agg_str: str, keep_order: bool = True) -> pl.DataFrame | Err:
    """
    Allows the following shorthands for `group_by`:
    - Use comma-delimited col names
    - Specify aggregators after `->` using list or dict syntax
        - For no aggregator specified, default to `.all()`

    Examples:
    - `"a"` -- `group_by('a').all()`
    - `"a, b"` -- `group_by(['a', 'b']).all()`
    - `"a -> ['*'.len()]"` -- `group_by('a').len()`
    - `"a -> ['b'.mean()]"` -- `group_by('a').agg([pl.col('b').mean()]))
    - `"a -> ['*'.mean()]"` -- `group_by('a').mean()`

    Supported aggregation functions:
    - `len()`
    - `sum()`, `mean()`
    - `max()`, `min()`, `median()`
    - `std()`, `var()`
    """
    # Parse `agg` str into halfs
    agg_str = agg_str.replace(" ", "")
    query_parts = agg_str.split("->")
    res = source
    if len(query_parts) == 1:
        # Default to using `.all()` aggregator
        col_names = agg_str.split(",")
        res = source.group_by(col_names, maintain_order=keep_order).all()
    elif len(query_parts) == 2:
        col_names = query_parts[0].split(",")
        res = source.group_by(col_names, maintain_order=keep_order)  # type: ignore
        agg_list_str = query_parts[1].removeprefix("[").removesuffix("]")
        agg_expr_list = _agg_list_to_polars_expr(agg_list_str)
        if isinstance(agg_expr_list, Err):
            return agg_expr_list
        else:
            res = res.agg(agg_expr_list)  # type: ignore
    else:
        raise ValueError("Groupby aggregation string contained too many `->` characters")

    if res.is_empty():
        return Err("Dataframe after `group_by` is empty")

    return res


def _check_assumptions(source: pl.DataFrame | Iterable[pl.DataFrame]) -> None:
    if isinstance(source, pl.DataFrame):
        source = (source,)
    for df in source:
        ## Check for column names that are `str`
        col_types = {type(c) for c in df.columns}
        if col_types != {str}:
            raise ValueError(f"Column headers need to be `str`, got: {col_types}")


def _pre_merge_checks(source: pl.DataFrame, second: pl.DataFrame, on: str | list[str]) -> None:
    # If _any_ of the provided indices aren't there, return `None`
    _check_assumptions([source, second])
    if isinstance(on, str):
        on = [on]
    for c in on:
        if not (c in source.columns and c in second.columns):
            raise KeyError(f"Proposed key {c} is not in either column!")


class PythonExprToPolarsExprVisitor(ast.NodeVisitor):
    def visit_BoolOp(self, node):
        if isinstance(node.op, ast.And):
            expr = self.visit(node.values[0])
            for value in node.values[1:]:
                expr = expr & self.visit(value)
        elif isinstance(node.op, ast.Or):
            expr = self.visit(node.values[0])
            for value in node.values[1:]:
                expr = expr | self.visit(value)
        return expr

    def visit_Compare(self, node):
        left = self.visit(node.left)
        comparisons = []
        for op, comparator in zip(node.ops, node.comparators):
            right = self.visit(comparator)
            if isinstance(op, ast.Eq):
                comparisons.append(left == right)
            elif isinstance(op, ast.Gt):
                comparisons.append(left > right)
            elif isinstance(op, ast.Lt):
                comparisons.append(left < right)
            elif isinstance(op, ast.GtE):
                comparisons.append(left >= right)
            elif isinstance(op, ast.LtE):
                comparisons.append(left <= right)
            elif isinstance(op, ast.NotEq):
                comparisons.append(left != right)
        expr = comparisons[0]
        for comparison in comparisons[1:]:
            expr = expr & comparison
        return expr

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        if isinstance(node.op, ast.Add):
            return left + right
        elif isinstance(node.op, ast.Sub):
            return left - right
        elif isinstance(node.op, ast.Mult):
            return left * right
        elif isinstance(node.op, ast.Div):
            return left / right
        elif isinstance(node.op, ast.Mod):
            return left % right

    def visit_Name(self, node):
        return pl.col(node.id)

    def visit_Constant(self, node):
        return pl.lit(node.value)

    def visit(self, node):
        return super().visit(node)


def _convert_to_polars_filter(filter_string: str) -> pl.Expr:
    tree = ast.parse(filter_string, mode="eval")
    visitor = PythonExprToPolarsExprVisitor()
    return visitor.visit(tree.body)


def _apply_nested_col_list(
    source: pl.DataFrame,
    parsed_nested_col_list: list[str | tuple[bool, list[str] | dict[str, str]]],
) -> pl.DataFrame:
    """
    Completes handling of the `.`, `->`, `+>` operators which is the `parsed_nested_col_list`.
      Converts the string expressions into corresponding Polars expression to apply at the end

    The incoming "parsed_nested_col_list" looks something like:
        ["some_col", "b", "c.d", (True, ["b", "c.d"]), (False, {"new_name": "c.d"}) ...]
    """

    # Handle "*" case -- replace each instance with `source.columns`
    if "*" in parsed_nested_col_list:
        # Find indices of all occurrences of "*"
        star_idx_list = [i for i, x in enumerate(parsed_nested_col_list) if x == "*"]
        # Replace each "*" with the replacement values
        for idx in reversed(star_idx_list):
            parsed_nested_col_list[idx : idx + 1] = source.columns

    # For each column specified, convert it to the corresponding Polars expression.
    #   Apply the expression at the end to get the final result
    res = source
    expr_list = []
    for i, nested_col in enumerate(parsed_nested_col_list):
        # Single-column case (e.g. "col_name")
        if isinstance(nested_col, str):
            nesting_expr = _nesting_to_polars_expr(nested_col)
            expr_list.append(nesting_expr)
        # # Expansion case (`->` or `+>`)
        elif isinstance(nested_col, tuple):
            keep_col, col_obj = nested_col
            # keep column if specified
            if keep_col:
                expr_list.append(pl.col(res.columns[i]))
            if isinstance(col_obj, list):
                for new_nesting in col_obj:
                    nesting_expr = _nesting_to_polars_expr(new_nesting)
                    expr_list.append(nesting_expr)
            # renaming case
            elif isinstance(col_obj, dict):
                for new_name, new_nesting in col_obj.items():
                    nesting_expr = _nesting_to_polars_expr(new_nesting, new_name)
                    expr_list.append(nesting_expr)
    if expr_list:
        res = res.select(expr_list)
    return res


def _nesting_to_polars_expr(nested_col: str, new_name: str | None = None) -> pl.Expr:
    """
    Converts something like:
        `a.b.c[0].d`
      into:
        `pl.col("a").struct.field("b").struct.field("c").list[0].struct.field("d")`
    """
    nesting_list = nested_col.split(".", maxsplit=1)

    res = pl.col(nesting_list[0])
    if len(nesting_list) > 1:
        for item in nesting_list[1].split("."):
            # Handle brackets -- grab value (assume just integer for now)
            list_idx: str | None = None
            if "[" in item:
                lbracket_idx = item.index("[")
                list_idx = item[lbracket_idx + 1 : item.index("]")]
                item = item[:lbracket_idx]  # Remove bracket from original str
            res = res.struct.field(item)
            if list_idx is not None:
                # TODO: Handle more than just single index, e.g. handle slices?
                res = res.list[int(list_idx)]

    if new_name:
        res = res.alias(new_name)
    else:
        res = res.alias(nested_col)

    return res


def _generate_nesting_list(
    parsed_col_list: list[str],
) -> list[str | tuple[bool, list[str] | dict[str, str]]]:
    """
    Return whether a specific column index should get nesting logic applied

    Given `parsed_col_list` as follows:
        ['c.d', 'reg_col', "some_json_col -> ['b', 'c.d']", "some_json_col +> {'new_name': 'c.d'}"]
      The resulting "parsed_nested_col_list" looks something like -- a tuple of `(keep_col, nesting)`:
        ['c.d', 'reg_col', (False, ['some_json_col.b', 'some_json_col.c.d']}), (True, {'new_name': 'some_json_col.c.d'})]
      A more complex nesting will index-into different values (set: one -> one/many new cols)
        and possibly rename them as well (dict: one -> one/many new cols with new names)

    For each column, check if:
      1. Column should be extracted and consumed (`->`)
      2. Column should be extracted and kept (`+>`)
      3. Column should be nested into and consumed (any other str and supporting `.` syntax)
    """
    parsed_nested_col_list: list[str | tuple[bool, list[str] | dict[str, str]]] = []

    for col_name in parsed_col_list:
        # 1. extract, and consume original
        # 2. extract, and keep original
        if ("->" in col_name) or ("+>" in col_name):
            keep_col = "+>" in col_name
            col_name, content = col_name.split("+>" if keep_col else "->")
            col_obj = _extract_list_or_dict(content, add_prefix=col_name)
            if col_obj:
                parsed_nested_col_list.append((keep_col, col_obj))
            else:
                raise RuntimeError("Error processing `->` or `+>` syntax")
        # 3. regular string nesting
        else:
            parsed_nested_col_list.append(col_name)
    return parsed_nested_col_list


def _extract_list_or_dict(
    s: str, add_prefix: str | None = None
) -> list[str] | dict[str, str] | None:
    """
    Tries to convert a string into a list[str] or dict[str, str]

    If `add_prefix` is specified, then adds the prefix to list/dict values (not keys)
    """
    try:
        # Clean string a bit
        s = s.replace(";", "").replace("(", "<fn").replace(")", ">")
        res = ast.literal_eval(s)
        if not (isinstance(res, list) or isinstance(res, dict)):
            raise ValueError("Need to specify a list or dict after `->` or `+>`")
        # Add a prefix to list/dict values if specified
        if add_prefix and isinstance(res, list):
            res = [f"{add_prefix}.{s}" for s in res]
        elif add_prefix and isinstance(res, dict):
            res = {k: f"{add_prefix}.{v}" for k, v in res.items()}
        return res
    except (ValueError, SyntaxError):
        return None


def _agg_list_to_polars_expr(agg_list_str: str) -> list[pl.Expr] | Err:
    """
    Takes something like:
        - "a -> ['*'.len()]"
        - "a -> ['b'.mean()]"
        - "a -> ['b'.mean(), 'c'.median()]
      and turns it into:
        - [pl.col('*').len()]
        - [pl.col('b').mean()]
        - [pl.col('b').mean(), pl.col('c').median()]

    Supported aggregation functions:
    - `len()`
    - `sum()`, `mean()`
    - `max()`, `min()`, `median()`
    - `std()`, `var()`
    """

    # Split into cols and aggregators
    # First split by commas to get each individual part
    # For each part:
    #  - Assume if `()` is present, it's an aggregator for the last noted col (in quotes)
    #  - Assume the first item must be a col name, and the first character must be a quote
    agg_cols = agg_list_str.split(",")
    agg_parts = []
    for col in agg_cols:
        col_parts = re.split(PERIOD_UP_TO_NEXT_CLOSE_PARENS, col)
        agg_parts.extend(col_parts)
    quote = agg_parts[0][0]
    if quote not in {'"', "'"}:
        return Err(f"Need to have column expressions in quotes, got `{agg_parts[0]}`")
    lexed_agg_list_str = str([p.strip(quote) for p in agg_parts if p != ""])
    parsed_agg_list_str = _extract_list_or_dict(
        lexed_agg_list_str
    )  # NOTE: this fn replaces `()` from the string with `<fn>`
    if parsed_agg_list_str is None:
        return Err(f"Could not parse expression `{agg_list_str}`")

    # Apply the appropriate aggregation function
    res = []
    curr_expr: pl.Expr | None = None
    for agg_part in parsed_agg_list_str:
        if not agg_part.endswith("<fn>"):
            # Assume default `all()` if nothing specified
            if curr_expr is not None:
                res.append(curr_expr.all())
            # Set-up next aggregator
            curr_expr = pl.col(agg_part)
        else:
            # NOTE: each of these aggregators will be considered "terminal". No chaining currently
            #   ... also no logic-checking / correcting. Just handling as-is!
            if not isinstance(curr_expr, pl.Expr):
                return Err(
                    f"Error when handling aggregation expression, tried to aggregate incorrect type: {curr_expr}"
                )
            match agg_part:
                case "len<fn>":
                    curr_expr = curr_expr.len()
                case "sum<fn>":
                    curr_expr = curr_expr.sum()
                case "mean<fn>":
                    curr_expr = curr_expr.mean()
                case "max<fn>":
                    curr_expr = curr_expr.max()
                case "min<fn>":
                    curr_expr = curr_expr.min()
                case "median<fn>":
                    curr_expr = curr_expr.median()
                case "std<fn>":
                    curr_expr = curr_expr.std()
                case "var<fn>":
                    curr_expr = curr_expr.var()  # .alias(f"{curr_expr.meta.output_name()}_sum")
                case _:
                    return Err(f"Got unsupported aggregator: {agg_part}")
            # Rename column based on the aggregator
            #   Handle the `*` case by checking for root name
            agg_suffix = agg_part.removesuffix("<fn>")
            if root_names := curr_expr.meta.root_names():
                curr_expr = curr_expr.alias(f"{root_names[0]}_{agg_suffix}")
            else:
                # The `*` case -- add a suffix
                curr_expr = curr_expr.name.suffix(f"_{agg_suffix}")
            res.append(curr_expr)
            curr_expr = None

    # Handle last item case (e.g. `b` for "a, b")
    if curr_expr is not None:
        res.append(curr_expr.all())

    return res



from copy import deepcopy
from typing import Any, Callable

import pydian.partials as p

from .rules import RC, RGC, Rule, RuleGroup


class IsRequired(Rule):
    """
    A rule where the current field is required

    For `RuleGroup`: keep default contraint
    """

    def __init__(self, at_key: str | None = None):
        # For each rule, make it required
        super().__init__(p.not_equivalent(None), RC.REQUIRED, at_key=at_key)

    def __and__(self, other: Rule | RuleGroup | Any, swap_order: bool = False) -> Rule | RuleGroup:
        """
        Returns a `RuleGroup` with the `REQUIRED` constraint applied to the single rule
        """
        match other:
            case Rule():
                res = deepcopy(other)
                res._constraint = RC.REQUIRED
            case _:
                # Check callable case here (cast into a `Rule`)
                if not isinstance(other, RuleGroup) and callable(other):
                    res = Rule.init_specific(other, RC.REQUIRED)  # type: ignore
                else:
                    if swap_order:
                        res = super().__rand__(other)
                    else:
                        res = super().__and__(other)
        return res

    def __rand__(self, other):
        return self.__and__(other, swap_order=True)


class IsOptional(Rule):
    """
    Optional: can be `None`, but validate if it's there.

    When combined with another rule, allows the case where the field is `None`.

    NOTE: This doesn't make sense to run on its own (unless you want a `None` check)
    """

    def __init__(self, at_key: str | None = None):
        # Initialize with dummy placeholder rule
        super().__init__(p.equivalent(None), at_key=at_key)

    def __and__(
        self, other: Rule | RuleGroup | Callable | type | Any, swap_order: bool = False
    ) -> Rule | RuleGroup:
        """
        Wraps in a `RuleGroup` where only 1 thing needs to pass (includes optional `None` check)
          This makes sure that condition is still run when data is present
        """
        # Wrap in a `RuleGroup` that has `None` equivalence as passing condition
        if callable(other):
            if swap_order:
                items = [Rule.init_specific(other), self]
            else:
                items = [self, Rule.init_specific(other)]
            res = RuleGroup(items, RGC.AT_LEAST_ONE)
        else:
            # Use `OR` here since we want `AT_LEAST_ONE` condition
            # ... this only happens for `IsOptional`... since it's special...
            res = super().__or__(other)
        return res

    def __rand__(self, other: Rule | RuleGroup | Any):
        return self.__and__(other, True)


class InRange(Rule):
    def __init__(
        self, lower: int | None = None, upper: int | None = None, at_key: str | None = None
    ):
        """
        Used to check if an list is within a size range, e.g.
            [
                str
            ] & InRange(3, 5)
          is a list of 3 to 5 `str` values

        """
        match (lower, upper):
            case (int(), None):
                fn = lambda l: len(l) >= lower
            case (None, int()):
                fn = lambda l: len(l) <= upper
            case (int(), int()):
                fn = lambda l: lower <= len(l) <= upper
            case (None, None):
                raise ValueError("Need to specify lower and/or upper bound: none received!")
        super().__init__(fn, at_key=at_key)


class MaxCount(Rule):
    def __init__(
        self,
        upper: int,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        super().__init__(p.pipe(len, p.lte(upper)), constraint, at_key)


class MinCount(Rule):
    def __init__(
        self,
        lower: int,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        super().__init__(p.pipe(len, p.gte(lower)), constraint, at_key)


class IsType(Rule):
    _type: type | None = None  # Store this for pydantic conversion

    def __init__(
        self,
        typ: type,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        self._type = typ
        super().__init__(p.isinstance_of(typ), constraint, at_key)


class InSet(Rule):
    """IDEA: have this be the enum variant. E.g. one of these literals"""

    def __init__(self, s: set[Any]):
        super().__init__(p.contained_in(s))



from __future__ import (  # Used to recursively type annotate (e.g. `Rule` in `class Rule`)
    annotations,
)

import inspect
from collections.abc import Callable, Collection, Iterable
from copy import deepcopy
from enum import Enum
from typing import Any

from result import Err, Ok

import pydian.partials as p

from ..dicts import get


class RC(Enum):
    """
    Rule Constraint (RC): A constraint for a single `Rule` (`RuleGroup` is responsible for application)

    Optional by default (i.e. no value specified)
    """

    REQUIRED = 1
    # ONLY_IF = lambda fn: (RC.REQUIRED, fn)  # Usage: RC.ONLY_IF(some_fn)
    # ONLY_AFTER = ... # Usage: ONLY_AFTER(other_rule) ... or some way to identify...


class RGC(Enum):
    """
    RuleGroup Constraint (RGC): Constraints on top of current `RuleGroup`

    Groups of >3 explicitly required are discouraged - KISS!

    RGC takes precedent over RC (i.e. a `RGC` setting will override a `RC` setting when applicable)
    """

    # TODO: Consider conditional feature -- `ONLY_IF(...)`, `ALL_IF(...)` instead of `ALL_WHEN_DATA_PRESENT`
    #       Would need to think through if they take generic callables, Rules, etc.

    ALL_REQUIRED_RULES = -2  # NOTE: This makes rules default _optional_
    ALL = -1  # NOTE: This makes rules default _required_
    ALL_WHEN_DATA_PRESENT = 0  # NOTE: This makes rules default _required when data is present_
    AT_LEAST_ONE = 1
    AT_LEAST_TWO = 2
    AT_LEAST_THREE = 3


class Rule:
    """
    A `Rule` is a callable that either returns:
    - Ok(truthy_result)
    - Err(str + falsy_result)

    A `Rule` can have additional constraints. These will ONLY be applied if contained within a `RuleGroup`

    `Rule`s can be combined to create `RuleGroup`s using: `&`, `|`
    """

    def _raise_undefined_rule_err():  # type: ignore
        raise ValueError("Rule was not defined!")

    _fn: Callable = lambda _: Rule._raise_undefined_rule_err()
    _constraint: RC | None = None
    _key: str | None = None
    _iter_over_input: bool | None = None

    def __init__(
        self,
        fn: Callable,
        constraint: RC | None = None,
        at_key: str | None = None,
    ):
        self._fn = fn
        self._key = at_key
        if constraint:
            self._constraint = constraint

    def __call__(
        self, source: Any, *args
    ) -> Ok[tuple[Rule, Any, Any]] | Err[tuple[Rule, Any, Any]]:
        """
        Call specified rule and wraps as a `Result` type.

        This will return one of the tuple variants of:
         1. Ok((current Rule, input, output))
         2. Err((current Rule, input, output))
         3. Err((current Rule, input, exception))
        """
        # NOTE: Only apply key logic for `dict`s. Something something, design choice!
        # Also: if passed an `Ok`, unwrap it by default
        if isinstance(source, dict) and self._key:
            curr_source = get(source, self._key)
        elif isinstance(source, Ok):
            curr_source = source.unwrap()
        else:
            curr_source = source
        try:
            # TODO: Handle list case (key = "[*]")
            # FYI: First case is we get explicitly want to process all items in list
            #  Second case is we assume we want to process all items in the list (based on parent)
            if self._iter_over_input:
                assert isinstance(curr_source, list), "Err: did not get a `list` for `[*]` key!"
                # Ok. We run the `_fn` for each item in the list, and return the results if all passed
                #   If there's any fail, then exit early, and the last item is a fail
                is_all_truthy = True
                res = []
                for it in curr_source:
                    it_res = self._fn(it)
                    res.append(it_res)
                    is_all_truthy = is_all_truthy and bool(it_res)
                    if not is_all_truthy:
                        raise RuntimeError(
                            f"Got failure when evaluating item {len(res)}: {res[-1]}"
                        )
            else:
                res = self._fn(curr_source, *args)
            if res:
                return Ok((self, source, res))
        except Exception as e:
            return Err((self, "ERROR", e))
        return Err((self, source, res))

    def __repr__(self) -> str:
        # NOTE: can only grab source for saved files, not in repl
        #  So the function needs to be saved on a file
        #  See: https://stackoverflow.com/a/335159
        try:
            return f"<Rule {self._fn.__name__} | {inspect.getsource(self._fn).strip()} | {tuple(c.cell_contents for c in self._fn.__closure__)}>"  # type: ignore
        except:
            try:
                return f"<Rule {self._fn.__name__} | {tuple(c.cell_contents for c in self._fn.__closure__)}>"  # type: ignore
            except:
                return f"<Rule {self._fn.__name__}>"

    def __hash__(self):
        return hash((self._fn, self._constraint, self._key))

    def __eq__(self, other: Rule | Any):
        if isinstance(other, Rule):
            # Rules are the same based on the code and the applied constraints
            # TODO: I _think_ this will work, though need to test more thoroughly
            #   e.g. this won't work for built-in callables like `str`, `bool`
            try:
                return other._fn.__code__.co_code == self._fn.__code__.co_code
            except:
                return other._fn == self._fn
        return NotImplemented

    @staticmethod
    def init_specific(
        v: Any, constraint: RC | None = None, at_key: str | None = None
    ) -> Rule | RuleGroup:
        """
        Generically returns a more specific rule when possible
        """
        # Don't re-wrap an existing Rule/RuleGroup
        if isinstance(v, Rule | RuleGroup):
            return v

        # Handle list and dict cases
        if isinstance(v, list):
            return _list_to_rulegroup(v)

        if isinstance(v, dict):
            return _dict_to_rulegroup(v)

        # TODO: consider `InRange` (take range obj) and `InSet` (take set object)
        from .specific import IsType  # Import here to avoid circular import

        if isinstance(v, type):
            res = IsType(v, constraint, at_key)  # type: ignore
        else:
            res = Rule(v, constraint, at_key)  # type: ignore
        return res

    def __and__(self, other: Rule | RuleGroup | Any):
        if _check_rand(self, other):
            return other.__rand__(self)

        return RuleGroup.combine(self, other)

    def __rand__(self, other: Rule | RuleGroup | Any):
        return NotImplemented

    def __or__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other, RGC.AT_LEAST_ONE)

    def __ror__(self, other: Rule | RuleGroup | Any):
        # Operation intended to be commutative
        return self.__or__(other)


class RuleGroup(list):
    """
    A `RuleGroup` is a callable list that can contain:
    - One or many `Rule`s
    - Other `RuleGroup`s (ending in a terminal `RuleGroup` of just `Rule`s at some point)

    The specified `RGC` (RuleGroup Constraint) determines the logic for the result outcome.
      Additionally, individual `Rule`s may have constraints which the `RuleGroup` manages
    """

    _constraint: RGC | None = None  # type: ignore
    _n_rules: int = 0
    # NOTE: _key is needed here to save nesting information from operations like `&` with a `dict`
    #   e.g. a user-specified `RuleGroup` shouldn't need to specify `_key` for each rule,
    #   rather we should infer that during parsing
    _key: str | None = None
    _iter_over_input: bool | None = None

    def __init__(
        self,
        items: Collection[Rule | RuleGroup | Callable] | Any | None = None,
        constraint: RGC = RGC.ALL,
        at_key: str | None = None,
    ):
        self._key = at_key
        self._constraint = constraint
        self._n_rules = 0

        # Type-check and handle items
        if items:
            res = []
            # Check items
            if not isinstance(items, Iterable):
                items = [items]
            for it in items:
                if not (isinstance(it, Rule) or isinstance(it, RuleGroup)):
                    # Add a new `Rule` wrapper if applicable
                    if callable(it):
                        it = Rule.init_specific(it)
                    else:
                        raise ValueError(
                            f"All items in a `RuleGroup` must be `Rule`s or `RuleGroup`s, got: {type(it)}"
                        )
                res.append(it)
                if isinstance(it, RuleGroup):
                    self._n_rules += it._n_rules
                else:
                    self._n_rules += 1
            super().__init__(res)
        else:
            super().__init__()

    def __deepcopy__(self, memo):
        """
        Custom deepcopy to prevent problem with `_n_rules` counting
        """
        # Create a new instance without calling __init__
        new_instance = RuleGroup.__new__(RuleGroup)
        memo[id(self)] = new_instance

        # Copy simple attributes
        new_instance._key = self._key
        new_instance._iter_over_input = self._iter_over_input
        new_instance._constraint = self._constraint
        new_instance._n_rules = self._n_rules

        # Deep copy the list items
        new_instance.extend(deepcopy(list(self), memo))

        return new_instance

    def append(self, item: Rule | RuleGroup | Callable):
        if not (isinstance(item, Rule) or isinstance(item, RuleGroup)):
            if callable(item):
                item = Rule.init_specific(item)
            else:
                raise ValueError(f"All items in a `RuleGroup` must be `Rule`s, got: {type(item)}")
        if isinstance(item, RuleGroup):
            self._n_rules += item._n_rules
        else:
            self._n_rules += 1
        super().append(item)

    def extend(self, item: RuleGroup | Iterable[Rule | RuleGroup | Callable]):
        if isinstance(item, RuleGroup):
            self._n_rules += item._n_rules
            # Copy-over key information if present (always override)
            if item._key:
                self._key = item._key
        super().extend(item)

    @staticmethod
    def combine(
        first: Rule | RuleGroup,
        other: Rule | RuleGroup | Any,
        constraint: RGC = RGC.ALL,
    ) -> RuleGroup:
        """
        Combines a `Rule | RuleGroup` with another value. By default, all data is optional by default

        Here are the expected cases:
        1. & (Rule | RuleGroup) -> Make a new RuleGroup with both existing ones (i.e. add a nesting parent)
        2. & type -> Add type check (IsType)
        3. & dict -> Add 1) type check, 2) contents of dict
        4. & list -> Add 1) type check, 2) contents of list (with `_iter_over_input`)
        5. & Callable -> Add the callable as a Rule
        6. & some primitive -> Add an equality check for the primitive
        """
        # Keep the nested structure if first is a RuleGroup
        it: tuple[RuleGroup] | Rule | RuleGroup = (
            (first,) if isinstance(first, RuleGroup) else first
        )
        res = RuleGroup(it, constraint)
        match other:
            case Rule() | RuleGroup():
                res.append(other)
            case dict():
                res.append(Rule.init_specific(dict))  # Type check
                drs = _dict_to_rulegroup(other)
                res.append(drs)
            case list():
                res.append(Rule.init_specific(list))  # Type check
                lrs = _list_to_rulegroup(other)
                res.append(lrs)
            case _:
                if callable(other):
                    res.append(Rule.init_specific(other))
                else:
                    res.append(Rule(p.equals(other)))
        return res

    def __call__(
        self, source: Any, *args
    ) -> Ok[tuple[RuleGroup, Any, RuleGroup]] | Err[tuple[RuleGroup, Any, RuleGroup]]:
        """
        Calls all embedded `Rule | RuleGroup`s and returns a corresponding result.

        This will return one of the tuple variants:
          - Ok((current RuleGroup, input, passed RuleGroup))
          - Err((current RuleGroup, input, failed RuleGroup))

        If the `RuleGroup` is nested: the returned `RuleGroup` will maintain the original structure.
        """
        # Apply key unnesting logic only when source is a dict. Design choice!
        if isinstance(source, dict) and self._key:
            curr_source = get(source, self._key)
        else:
            curr_source = source

        # Run each rule and save results
        # NOTE: This nests results in a RuleGroup by default. For recursive calls, we'll unnest this below
        rg_passed, rg_failed = RuleGroup(constraint=RGC.ALL), RuleGroup(constraint=RGC.ALL)
        for curr_item in self:
            assert isinstance(
                curr_item, Rule | RuleGroup
            ), f"Expected <Rule | RuleGroup> whe calling RuleGroup, got {type(curr_item)}"

            # Pass-down parent iter info
            if isinstance(curr_source, list) and self._iter_over_input:
                curr_item._iter_over_input = True

            # Run the rule(s)
            if self._iter_over_input and isinstance(curr_source, list):
                # Ok. We run the `_fn` for each item in the list, and return the results if all passed
                #   If there's any fail, then exit early, and the last item is a fail
                is_all_truthy = True
                iter_list = []
                for it in curr_source:
                    it_res = curr_item(it)
                    iter_list.append(it_res)
                    is_all_truthy = is_all_truthy and bool(it_res)
                    if not is_all_truthy:
                        raise RuntimeError(
                            f"Got failure when evaluating item {len(iter_list)}: {iter_list[-1]}"
                        )
                curr_res = Ok(("[*]", curr_source, curr_item))
            else:
                curr_res = curr_item(curr_source, *args)  # type: ignore
            # Save result based on cases
            match curr_item, curr_res:
                case (Rule(), Ok()):
                    rg_passed.append(curr_item)
                case (Rule(), Err()):
                    rg_failed.append(curr_item)
                # For RuleGroup: take last item in Ok/Err tuple
                case (RuleGroup(), Ok()):
                    rg_passed.append(curr_res.ok_value[-1])  # type: ignore
                case (RuleGroup(), Err()):
                    rg_failed.append(curr_res.err_value[-1])  # type: ignore
                case _:
                    raise RuntimeError(
                        f"Unexpected type or result: {type(curr_item)}, {type(curr_res)}"
                    )

        ## Check for failed required rules -- return Err early if so
        if _contains_required_rule(rg_failed):
            return Err((self, source, rg_failed))

        # Check result and return
        res: Ok[tuple[RuleGroup, Any, RuleGroup]] | Err[tuple[RuleGroup, Any, RuleGroup]]
        passed_case = (self, source, rg_passed)
        failed_case = (self, source, rg_failed)
        match self._constraint:
            case RGC.ALL:
                res = Ok(passed_case) if len(rg_passed) == len(self) else Err(failed_case)
            case RGC.AT_LEAST_ONE | RGC.AT_LEAST_TWO | RGC.AT_LEAST_THREE:
                res = (
                    Ok(passed_case)
                    if len(rg_passed) >= self._constraint.value
                    else Err(failed_case)
                )
            case RGC.ALL_REQUIRED_RULES:
                # Since we have above check for required rules, we know all rules have passed here
                res = Ok(passed_case)
            case RGC.ALL_WHEN_DATA_PRESENT:
                # For each failed rule, check if data was present. If so, return `Err`
                res = (
                    Ok(passed_case)
                    if not _rulegroup_applies(rg_failed, curr_source)
                    else Err(failed_case)
                )
            case _:
                raise RuntimeError(f"Unsupported RuleGroup constraint: {self._constraint}")
        return res

    def __hash__(self):
        return hash(tuple(self))

    def __repr__(self) -> str:
        # TODO: Consider something less verbose
        return f"<RuleGroup len={len(self)} n={self._n_rules} {self._constraint}>"

    def __and__(self, other: Rule | RuleGroup | Any):
        if _check_rand(self, other):
            return other.__rand__(self)
        return RuleGroup.combine(self, other)

    def __rand__(self, other: Rule | RuleGroup | Any):
        # Expect commutative
        return NotImplemented

    def __or__(self, other: Rule | RuleGroup | Any):
        return RuleGroup.combine(self, other, RGC.AT_LEAST_ONE)

    def __ror__(self, other: Rule | RuleGroup | Any):
        # Expect commutative
        return self.__or__(other)


""" Helper Functions """


def _check_rand(curr: Rule | RuleGroup, other: Rule | RuleGroup | Any):
    """
    Checks if there's a more specific `__and__` to call (`rand` is "right and" in this context)
    """
    try:
        if hasattr(other, "__rand__") and other.__rand__(curr) is not NotImplemented:
            return True
    except:
        return False


def _contains_required_rule(rg: RuleGroup | Rule) -> bool:
    """
    Returns `True` if `rg` contains at least 1 Rule with required constraint
    """
    if isinstance(rg, Rule):
        return rg._constraint == RC.REQUIRED
    return any(_contains_required_rule(r) for r in rg)


def _rulegroup_applies(rg: RuleGroup | Rule, source: dict[str, Any]) -> bool:
    """
    Returns `True` if `rg` applies to _any_ part of `source`
      This is mainly to handle the `ALL_WHEN_DATA_PRESENT` logic

    If there is no key-level data in `rg`, then conservatively assume there is overlap
    """
    if isinstance(rg, Rule):
        return rg._key is None or get(source, rg._key) is not None
    return any(_rulegroup_applies(r, source) for r in rg)


def _list_to_rulegroup(
    l: list[Callable | Rule | RuleGroup | dict | list], key_prefix: str | None = None
) -> RuleGroup:
    """
    Given a list, compress it into a single `RuleGroup`
      (or for the `[*]` case: a `Rule` containing a single `RuleGroup`)

    An item in the list should pass at least one of the callables in the `RuleGroup`.
      For example: `{ 'k': [ str, bool ]}` -- at key 'k', it can contain a list `str` or `bool`
      (for more specific constraints: use a nested `RuleGroup`)
    """
    # NOTE: by default, elements in a list are considered joined by `AND` logic
    res = RuleGroup(constraint=RGC.ALL, at_key=key_prefix)
    # NOTE: by default, calling this will always add a `[*]`
    #   to make sure rules are applied over the list
    res._iter_over_input = True
    for it in l:
        # TODO: does this even work as-expected? Not sure for atomic `Rule`s...
        match it:
            case Rule() | RuleGroup():
                it._iter_over_input = True
                if isinstance(it, Rule):
                    res.append(it)
                else:
                    # In this case, avoid the double-nesting and just overwrite key info
                    res = it
            case dict():
                res.extend(_dict_to_rulegroup(it))
            case list():
                res.extend(_list_to_rulegroup(it))
            case _:
                if callable(it):
                    new_rule = Rule.init_specific(it)
                else:
                    new_rule = Rule(p.equals(it))  # Exact value check
                new_rule._iter_over_input = True
                res.append(new_rule)
    return res


def _dict_to_rulegroup(d: dict[str, Rule | RuleGroup], key_prefix: str | None = None) -> RuleGroup:
    """
    Given a dict, compress it into a single `RuleGroup` with key information saved

    NOTE: expect this dict to be 1-layer (i.e. not contain other dicts) -- other dicts should
      be encompassed in a `RuleGroup`
    """
    res = RuleGroup(at_key=key_prefix)
    for k, v in d.items():
        match v:
            case Rule() | RuleGroup():
                v._key = k
                res.append(v)
            case dict():
                res.append(Rule.init_specific(dict, at_key=k))
                res.append(_dict_to_rulegroup(v, key_prefix=k))
            case list():
                res.append(Rule.init_specific(list, at_key=k))
                res.append(_list_to_rulegroup(v, key_prefix=k))
            case _:
                if callable(v):
                    res.append(Rule.init_specific(v, at_key=k))
                else:
                    # Exact value check
                    res.append(Rule(p.equals(v), at_key=k))
    return res



import traceback
from typing import Any

from .dicts import drop_keys, impute_enum_values
from .globs import SharedMapperState, _Global_Mapper_State_Dict
from .lib.types import DROP, KEEP, MappingFunc
from .lib.util import encode_stack_trace, get_keys_containing_class, remove_empty_values


class Mapper:
    def __init__(
        self,
        map_fn: MappingFunc,
        remove_empty: bool = True,
        strict: bool = False,
    ) -> None:
        self.map_fn = map_fn
        self.remove_empty = remove_empty
        self.strict = strict
        self.global_mapper_call_id: str | None = None
        self.global_mapper_call_level: int | None = None

    def _register_mapper_call_id(self) -> None:
        """
        A `mapper_call_id` is uniquely identified by the current stack trace.
          This is so subsequence function calls can identify it is part of the mapper.
        """
        # The stack trace, excluding this level and `__call__`
        curr_trace = traceback.format_stack()[:-2]
        self.global_mapper_call_id = encode_stack_trace(curr_trace)
        self.global_mapper_call_level = len(curr_trace)
        # Update global state
        _Global_Mapper_State_Dict[self.global_mapper_call_id] = SharedMapperState(
            _trace_len=self.global_mapper_call_level,
            strict=self.strict,
        )
        return None

    def __call__(self, source: dict[str, Any], **kwargs) -> dict[str, Any]:
        """
        Calls `map_fn` and then performs postprocessing into the result dict.
        """
        # If global mapper ID is not in config dict, then add it + hash stack trace
        if not self.global_mapper_call_id:
            self._register_mapper_call_id()

        res = self.map_fn(source, **kwargs)

        # Handle any DROP-flagged values
        keys_to_drop = get_keys_containing_class(res, DROP)
        if keys_to_drop:
            res = drop_keys(res, keys_to_drop)

        # Remove empty values
        if self.remove_empty:
            res = remove_empty_values(res)

        # Impute KEEP values with corresponding value
        keys_to_impute = get_keys_containing_class(res, KEEP)
        if keys_to_impute:
            res = impute_enum_values(res, keys_to_impute)

        return res
